"""
Formula parsing

"""



import sys
from pyparsing import *

from typing import TypeAlias

from pysmt.shortcuts import (TRUE, FALSE, And, Or, Not, Implies, Iff, Equals)
from pysmt.shortcuts import Symbol
from mc308.data_structures.formula import PropFormula
from mc308.data_structures.ltl import LTLU,LTLX,LTLG,LTLF



# PARSING
ParserElement.enablePackrat()
sys.setrecursionlimit(3000)

TRUETK = Keyword("True")
FALSETK = Keyword("False")


start_chars = alphas + "_"
ltlop = ["X","G","U","F"]
for op in ltlop:
    start_chars = start_chars.replace(op,"")
identifier_chars = alphanums
print(identifier_chars)
print("")
identifier = Word(start_chars, bodyChars=identifier_chars, exact=1)# - excluded_words

operand = TRUETK | FALSETK | identifier

notop = Keyword("!")
andop = Keyword("&")
orop = Keyword("|")
impliesop = Keyword("->")
iffop =  Keyword("<->")

nextop =  Keyword("X")
globallyop =  Keyword("G")
finallyop =  Keyword("F")
untilop =  Keyword("U")




TRUETK.setParseAction(lambda t : TRUE())
FALSETK.setParseAction(lambda t : FALSE()) 

identifier.setParseAction(lambda t : Symbol(t[0]))


class ParseBexp():
    """
    Precedence: !, X, F ,G
    Unary > binary
    U precedence over and, or, implies, iff
    """
    def __init__(self):
        self.parser = infixNotation(
            operand,
            [
                (notop, 1, opAssoc.RIGHT, lambda t: Not(t[0][1])),
                # LTL
                (nextop, 1, opAssoc.RIGHT, lambda t : LTLX(t[0][1])),
                (finallyop, 1, opAssoc.RIGHT, lambda t : LTLF(t[0][1])),
                (globallyop, 1, opAssoc.RIGHT, lambda t : LTLG(t[0][1])),
                # UNTIL
                (untilop, 2, opAssoc.LEFT, lambda t : LTLU(t[0][0],t[0][2])),
                # BOOL
                (andop, 2, opAssoc.LEFT, lambda t : And(t[0][0],t[0][2])),
                (orop, 2, opAssoc.LEFT, lambda t : Or(t[0][0],t[0][2])),
                (impliesop, 2, opAssoc.LEFT, lambda t : Implies(t[0][0],t[0][2])),
                (iffop, 2, opAssoc.LEFT, lambda t : Iff(t[0][0],t[0][2])),

            ],
        )

    def parse(self, string):
        res = self.parser.parseString(string)
        return res[0]

 
