""" Module for printing the NFA.

TODO: reduce code duplication with the printing of TS (here, copy and paste)

Uses the graphviz library to create a DOT file.
"""

import io
import typing

import graphviz

from mc308.data_structures.nba import NBA, GNBA

from functools import cmp_to_key

class DotConfig:
    """
    Configuration for printing in dot format (now empty)
    """
    def __init__(self):
        pass

def nba_to_dot(nba : NBA | GNBA, config : DotConfig = DotConfig()) -> graphviz.Graph:
    """ Convert a transition system to a graph in the dot format """

    dot = graphviz.Digraph(comment='NBA')

    is_gnba = isinstance(nba, GNBA)


    # Create nodes
    for n in sorted(list(nba.states)):


        if not is_gnba:
            if n in nba.final:
                dot.node('%s' % str(n), shape='doublecircle')
            else:
                dot.node('%s' % str(n))
        else:
            for f in nba.final:
                if n in f:
                    dot.node('%s' % str(n), shape='doublecircle')
                else:
                    dot.node('%s' % str(n))

        # Add initial states
        if n in sorted(list(nba.init)):
            invisible_node = '_fake_init_%s' % str(n)
            dot.node(invisible_node, style="invisible")
            dot.edge('%s' % invisible_node, '%s' % str(n))

    # Create the edges
    def edge_str(a):
        return "%s %s %s" % (str(a[0]), str(a[2]), str(a[1]))

    for a in sorted(nba.tr, key = cmp_to_key(lambda x,y: edge_str(x) < edge_str(y))):
        dot.edge(str(a[0]), str(a[2]), str(a[1]))

    return dot



