import logging
import os
import sys


from . import TestCase

from pyparsing.exceptions import ParseException
from mc308.data_structures.formula import PropFormula
from mc308.data_structures.formula_parsing import ParseBexp
from mc308.data_structures.ltl import *
from mc308.data_structures.ltl_functions import *
from mc308.data_structures.ltl2buchi import *


from typing import Set, List, Collection


class TestLTL(TestCase):
    def test_normal_form(self):
        parser = ParseBexp()
        test_cases = [
            ("a","a"),
            ("TRUE", "TRUE"),
            ("! a", "! a"),
            ("! ! a", "! ! a"),
            ("a & b", "a & b"),
            ("a | b", "(! ((! a) & (! b)))"),
            ("a -> b", "(! (a & (! b)))"),
            ("a <-> (c -> b)", "((! (a & (c & (! b)))) & (! ((! (c & (! b))) & (! a))))"),
            ("G a", "(! (True U (! a)))"),
            ("F a & b", "((True U a) & b)"),
            ("X ! a", "(X (! a))"),
            ("F (a U b)", "(True U (a U b))"),
            ("a & (! ((a | b) U b))","(a & (! ((! ((! a) & (! b))) U b)))"),
        ]

        for t,expected in test_cases:
            res = normalize_ltl(parser.parse(t))
            expected_res = parser.parse(expected)
            self.assertTrue(res == expected_res)

    def test_subs(self):
        parser = ParseBexp()
        test_cases = [
            ("a",["a"]),
            ("a & b", ["(a & b)","a","b"]),
            ("! a", ["(! a)", "a"]),
            ("a | b", ["(a | b)", "a", "b"]),
            ("G a", ["(G a)", "a"]),
            ("F a & b", ["(F a) & b)", "a", "b", "F a"]),
            ("X ! a", ["X ! a", "! a", "a"]),
            ("F (a U b)",["F (a U b)", "a", "b", "a U b" ]),

            ("a & (! ((a | b) U b))",["a & (! ((a | b) U b))",
                                      "a",
                                      "! ((a | b) U b)",
                                      "((a | b) U b)",
                                      "(a | b)",
                                      "b"
                                      ])
        ]

        for t,expected in test_cases:
            res = parser.parse(t)
            subs = get_sub(res)
            self.assertTrue(set(subs) == set([parser.parse(f) for f in expected]))


    def test_elementary(self):
        def eq_lists(list1, list2):
            # Flatten the lists of lists and convert them to sets
            set1 = set(frozenset([item for item in sublist]) for sublist in list1)
            set2 = set(frozenset([item for item in sublist]) for sublist in list2)

            # Check if the sets are equal
            return set1 == set2

        def parse_expected(exp):
            # parse expected output, so that you have formulas and not strings
            for a in exp:
                for p in a:
                    return [[parser.parse(a) for a in sublist] for sublist in exp]

        parser = ParseBexp()
        test_cases = [
            ("a", [["a"],["! a"]]),
            ("! a", [["a"],["! a"]]),
            ("a U (! a & b)",
             [
                 ["a", "b","!(! a & b)","!(a U (! a & b))"],
                 ["a", "b","!(! a & b)", "(a U (! a & b))"],
                 ["a","! b","!(! a & b)", "(a U (! a & b))"],
                 ["a","! b","!(! a & b)","!(a U (! a & b))"],
                 ["! a","! b","!(! a & b)","!(a U (! a & b))"],
                 ["! a", "b","(! a & b)", "(a U (! a & b))"],
             ]
            ),

            ("X a", [
                ["a","X a"],
                ["a","! (X a)"],
                ["! a","X a"],
                ["! a","! (X a)"]
            ]
            )
        ]

        # Run all the tests
        for t,expected in test_cases:
            cls = closure(parser.parse(t))
            elementary = get_elementary(cls)
            res_expected = parse_expected(expected)

            # the elmentary set must be equal to the expected result
            self.assertTrue(eq_lists(elementary, res_expected))



class TestLTL2Buchi(TestCase):

    def test_Xa(self):
        parser = ParseBexp()
        ltlformula = parser.parse("X a")
        gnba = ltl2buchi(ltl2buchi)

        # You need to implement the comparison!
        raise NotImplementedError()


    def test_aUb(self):
        parser = ParseBexp()
        ltlformula = parser.parse("a U b")
        gnba = ltl2buchi(ltl2buchi)

        # You need to implement the comparison!
        raise NotImplementedError()

    def test_XU(self):
        parser = ParseBexp()
        ltlformula = parser.parse("(a & X a) U (a & ! X a)")
        gnba = ltl2buchi(ltl2buchi)

        # You need to implement the comparison!
        raise NotImplementedError()



